/*
 * FreeRTOS Kernel V10.2.0
 * Copyright (C) 2019 Amazon.com, Inc. or its affiliates.  All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * http://www.FreeRTOS.org
 * http://aws.amazon.com/freertos
 *
 * 1 tab == 4 spaces!
 */


#include "system.h"

#include "FreeRTOS.h"
#include "task.h"

#include "nios2.h"
#include "altera_vic_irq.h"
#include "altera_vic_regs.h"
#include "altera_avalon_timer_regs.h"
#include "altera_avalon_pio_regs.h"

#if ALT_NUM_EXTERNAL_INTERRUPT_CONTROLLERS < 1
	#error "No VIC configured"
#elif ALT_NUM_EXTERNAL_INTERRUPT_CONTROLLERS > 1
	#error "Only one VIC is allowed in this implementation!"
#endif

#ifdef ALT_MODULE_CLASS_sys_clk_timer
	#if SYS_CLK_TIMER_IRQ != 0
		#error "SYS timer for the RTOS tick has to be IRQ0!"
	#endif
#else
	#error "SYS timer for the RTOS not defined!"
#endif


/* Function prototype for the rtos system tick and yield handler */
void vPortSysTickHandler( void * context);

/* Status control register init value for the pxPortInitialiseStack
 * method. Bit 1 is the interrupt enable bit, bit 23 is RSIE (register
 * set interrupt-enable bit) to enable shadow registers */
#define portINITIAL_STATUS     ( StackType_t ) ((1 << NIOS2_STATUS_RSIE_OFST) | \
											    (1 << NIOS2_STATUS_PIE_OFST) )

static void prvReadGp( uint32_t *ulValue )
{
	asm( "stw gp, (%0)" :: "r"(ulValue) );
}

/*  Comment from tasks.c, line 1007:
    Initialize the TCB stack to look as if the task was already running,
	but had been interrupted by the scheduler.  The return address is set
	to the start of the task function. Once the stack has been initialized
	the top of stack variable is updated. */
StackType_t *pxPortInitialiseStack( StackType_t *pxTopOfStack, TaskFunction_t pxCode, void *pvParameters )
{
	StackType_t *pxFramePointer = pxTopOfStack - 1;
	StackType_t xGlobalPointer;

	prvReadGp( &xGlobalPointer );

	/* End of stack marker. */
	*pxTopOfStack = 0xdeadbeef;
	pxTopOfStack--;

	*pxTopOfStack = ( StackType_t ) pxFramePointer;
	pxTopOfStack--;

	*pxTopOfStack = xGlobalPointer;

	/* Space for R23 to R16. */
	pxTopOfStack -= 9;

	*pxTopOfStack = ( StackType_t ) pxCode;
	pxTopOfStack--;

	*pxTopOfStack = portINITIAL_STATUS;

	/* Space for R15 to R5. */
	pxTopOfStack -= 12;

	*pxTopOfStack = ( StackType_t ) pvParameters;

	/* Space for R3 to R1, muldiv and RA. */
	pxTopOfStack -= 5;

	// reserved
	pxTopOfStack--; 	

	// reserved
	pxTopOfStack--;

	return (pxTopOfStack);
}

BaseType_t xPortStartScheduler( void )
{
	/*
	 * ISR for RTOS sys timer without interrupt funnel into VIC vector table (altera_uC_VIC_vector_tbl.S):
	 * 	         jmpi isr_handler
	 *           .word   0x0	# alignment to 16 Byte
	 *           .word   0x0	# alignment to 16 Byte
	 *           .word   0x0	# alignment to 16 Byte
	 * The altera_uC_VIC_vector_tbl.S will be overwritten when generating the BSP
	 *
	 * The PIE bit and the IL field (interrupt level) will be set to true inside the first context switch by
	 * writing predefined value in estatus.
	 *
	 * Enable timer interrupt inside timer peripheral (ALTERA_AVALON_TIMER_CONTROL_ITO_MSK) and the
	 * corresponding VIC_ENABLE bit. The first interrupt will occure after restoring the first task context
	 * as soon the PIE bit will be set (last eret call will copy estatus to status)
	 *
	 * */
	IOWR_ALTERA_AVALON_TIMER_CONTROL(SYS_CLK_TIMER_BASE, ALTERA_AVALON_TIMER_CONTROL_ITO_MSK);
	IOWR_ALTERA_VIC_INT_ENABLE_SET(alt_vic_dev_list[0]->base, 1);

	/* Start the first task. */
    asm volatile (  " movia r2, restore_sp_from_pxCurrentTCB        \n"
                    " jmp r2                                          " );

	/* Should not get here! */
	return (0);
}

void vPortEndScheduler( void )
{
	/* It is unlikely that the NIOS2 port will require this function as there
	is nothing to return to.  */
}

//--------------------

void vPortSysTickHandler( void * context)
{
	/* Handle the system tick interrupt and the portYIELD_FROM_ISR exceptions.
	 * SysTick: Interrupt which jumps directly to "isr_handler"
	 * portYIELD_FROM_ISR: executes a software interrupt for the SysTick interrupt, which
	 *                     causes the same jump to "isr_handler" after the calling ISR is left
	 */

	// check if the SysTick timer has caused the exception
	if(IORD_ALTERA_AVALON_TIMER_STATUS(SYS_CLK_TIMER_BASE) & ALTERA_AVALON_TIMER_STATUS_TO_MSK){
		/* Increment the kernel tick. */
		if( xTaskIncrementTick() != pdFALSE )
		{
			vTaskSwitchContext();
		}

		/* Clear the interrupt. */
		IOWR_ALTERA_AVALON_TIMER_STATUS(SYS_CLK_TIMER_BASE, 0 );
	}else{
		// software interrupt, just call the context switch
		vTaskSwitchContext();
	}
	// make sure SW IRQ is cleared!
	IOWR_ALTERA_VIC_SW_INTERRUPT_CLR(alt_vic_dev_list[0]->base, 1);
}


void vPortDisableInterrupts( void )
{
	register uint32_t tmpStatus;
	NIOS2_READ_STATUS(tmpStatus);
	tmpStatus &= ~NIOS2_STATUS_IL_MSK;
	NIOS2_WRITE_STATUS(tmpStatus | (configMAX_SYSCALL_INTERRUPT_PRIORITY << NIOS2_STATUS_IL_OFST) );
}

void vPortEnableInterrupts( void )
{
	register uint32_t tmpStatus;
	NIOS2_READ_STATUS(tmpStatus);
	NIOS2_WRITE_STATUS(tmpStatus & (~NIOS2_STATUS_IL_MSK) );
}

UBaseType_t vPortSetInterruptMaskFromISR(void){
	register uint32_t tmpStatus;
	register uint32_t tmpStatutsILField;

	NIOS2_READ_STATUS(tmpStatus);
	tmpStatutsILField = tmpStatus & NIOS2_STATUS_IL_MSK;
	tmpStatus &= ~NIOS2_STATUS_IL_MSK; //set IL field to 0
	NIOS2_WRITE_STATUS( tmpStatus| (configMAX_SYSCALL_INTERRUPT_PRIORITY << NIOS2_STATUS_IL_OFST) );

	return tmpStatutsILField;
}

void vPortClearInterruptMaskFromISR(UBaseType_t uxSavedInterruptStatus){
	register uint32_t tmpStatus;

	NIOS2_READ_STATUS(tmpStatus);
	tmpStatus &= ~NIOS2_STATUS_IL_MSK; //set IL field to 0
	NIOS2_WRITE_STATUS(tmpStatus | uxSavedInterruptStatus );
}

